@@ -1,8 +1,12 @@
-my class Bool {
+my class Bool { # declared in BOOTSTRAP
+    # class Bool is Cool {
+    #     has int $!value;
+
     multi method Bool(Bool:D:)    { self }
     multi method Numeric(Bool:D:) { self ?? 1 !! 0 }
     multi method Str(Bool:D:)     { self ?? 'True' !! 'False' }
     multi method gist(Bool:D:)    { self ?? 'True' !! 'False' }
+    multi method DUMP(Bool:D:)    { self.Str }
 
     method Int()     { self ?? 1 !! 0 }
 
@@ -58,7 +62,7 @@ multi infix:<?|>(Mu \a, Mu \b)        { a.Bool || b.Bool }
 
 proto infix:<?^>(|) is pure           { * }
 multi infix:<?^>(Mu $x = Bool::False) { $x.Bool }
-multi infix:<?^>(Mu \a, Mu \b)        { a.Bool ^^ b.Bool }
+multi infix:<?^>(Mu \a, Mu \b)        { nqp::p6bool(nqp::ifnull(nqp::xor(a.Bool,b.Bool), 0)) }
 
 # These operators are normally handled as macros in the compiler;
 # we define them here for use as arguments to functions.
