@@ -21,16 +21,16 @@ my class Exception {
     }
 
     method throw() is hidden_from_backtrace {
-        nqp::bindattr(self, Exception, '$!ex', pir::new__Ps('Exception'))
-            unless nqp::defined($!ex);
-        pir::setattribute__vPsP($!ex, 'payload', nqp::p6decont(self));
+        nqp::bindattr(self, Exception, '$!ex', nqp::newexception())
+            unless nqp::isconcrete($!ex);
+        nqp::setpayload($!ex, nqp::decont(self));
         my $msg := self.?message;
-        pir::setattribute__0PsP($!ex, 'message', nqp::unbox_s($msg.Str))
+        nqp::setmessage($!ex, nqp::unbox_s($msg.Str))
             if $msg.defined;
         nqp::throw($!ex)
     }
     method rethrow() is hidden_from_backtrace {
-        pir::setattribute__vPsP($!ex, 'payload', nqp::p6decont(self));
+        nqp::setpayload($!ex, nqp::decont(self));
         nqp::rethrow($!ex)
     }
 
@@ -51,7 +51,7 @@ my class Exception {
     method fail(Exception:D:) {
         try self.throw;
         my $fail := Failure.new($!);
-        my Mu $return := pir::find_caller_lex__Ps('RETURN');
+        my Mu $return := nqp::getlexcaller('RETURN');
         $return($fail) unless nqp::isnull($return);
         $fail
     }
@@ -88,16 +88,21 @@ my class X::Method::InvalidQualifier is Exception {
 
 
 sub EXCEPTION(|) {
-    my Mu $parrot_ex := nqp::shift(pir::perl6_current_args_rpa__P());
-    my Mu $payload   := nqp::atkey($parrot_ex, 'payload');
+    my Mu $vm_ex   := nqp::shift(nqp::p6argvmarray());
+    my Mu $payload := nqp::getpayload($vm_ex);
     if nqp::p6bool(nqp::istype($payload, Exception)) {
-        nqp::bindattr($payload, Exception, '$!ex', $parrot_ex);
+        nqp::bindattr($payload, Exception, '$!ex', $vm_ex);
         $payload;
     } else {
-        my int $type = nqp::atkey_i($parrot_ex, 'type');
+        my int $type = nqp::getextype($vm_ex);
         my $ex;
-        if $type == pir::const::EXCEPTION_METHOD_NOT_FOUND  &&
-            nqp::p6box_s(nqp::atkey_s($parrot_ex, 'message')) ~~ /"Method '" (.*?) "' not found for invocant of class '" (.+)\'$/ {
+#?if parrot
+        if $type == pir::const::EXCEPTION_METHOD_NOT_FOUND &&
+#?endif
+#?if !parrot
+        if
+#?endif
+            nqp::p6box_s(nqp::getmessage($vm_ex)) ~~ /"Method '" (.*?) "' not found for invocant of class '" (.+)\'$/ {
 
             $ex := X::Method::NotFound.new(
                 method   => ~$0,
@@ -105,89 +110,40 @@ sub EXCEPTION(|) {
             );
         }
         else {
+
             $ex := nqp::create(X::AdHoc);
-            nqp::bindattr($ex, X::AdHoc, '$!payload', nqp::p6box_s(nqp::atkey_s($parrot_ex, 'message')));
+            nqp::bindattr($ex, X::AdHoc, '$!payload', nqp::p6box_s(nqp::getmessage($vm_ex)));
         }
-        nqp::bindattr($ex, Exception, '$!ex', $parrot_ex);
+        nqp::bindattr($ex, Exception, '$!ex', $vm_ex);
         $ex;
     }
 }
 
 my class X::Comp::AdHoc { ... }
 sub COMP_EXCEPTION(|) {
-    my Mu $parrot_ex := nqp::shift(pir::perl6_current_args_rpa__P());
-    my Mu $payload   := nqp::atkey($parrot_ex, 'payload');
+    my Mu $vm_ex   := nqp::shift(nqp::p6argvmarray());
+    my Mu $payload := nqp::getpayload($vm_ex);
     if nqp::p6bool(nqp::istype($payload, Exception)) {
-        nqp::bindattr($payload, Exception, '$!ex', $parrot_ex);
+        nqp::bindattr($payload, Exception, '$!ex', $vm_ex);
         $payload;
     } else {
         my $ex := nqp::create(X::Comp::AdHoc);
-        nqp::bindattr($ex, Exception, '$!ex', $parrot_ex);
-        nqp::bindattr($ex, X::AdHoc, '$!payload', nqp::p6box_s(nqp::atkey_s($parrot_ex, 'message')));
+        nqp::bindattr($ex, Exception, '$!ex', $vm_ex);
+        nqp::bindattr($ex, X::AdHoc, '$!payload', nqp::p6box_s(nqp::getmessage($vm_ex)));
         $ex;
     }
 }
 
-my class X::Comp::Group is Exception {
-    has $.panic;
-    has @.sorrows;
-    has @.worries;
-    
-    method is-compile-time() { True }
-    
-    method gist(::?CLASS:D:) {
-        my $r = "";
-        if $.panic || @.sorrows {
-            my $color = %*ENV<RAKUDO_ERROR_COLOR> // $*OS ne 'MSWin32';
-            my ($red, $clear) = $color ?? ("\e[31m", "\e[0m") !! ("", "");
-            $r ~= "$red==={$clear}SORRY!$red===$clear\n";
-            for @.sorrows {
-                $r ~= .gist(:!sorry, :!expect) ~ "\n";
-            }
-            if $.panic {
-                $r ~= $.panic.gist(:!sorry) ~ "\n";
-            }
-        }
-        if @.worries {
-            $r ~= $.panic || @.sorrows
-                ?? "Other potential difficulties:\n"
-                !! "Potential difficulties:\n";
-            for @.worries {
-                $r ~= .gist(:!sorry, :!expect).indent(4) ~ "\n";
-            }
-        }
-        $r
-    }
-    
-    method message() {
-        my @m;
-        for @.sorrows {
-            @m.push(.message);
-        }
-        if $.panic {
-            @m.push($.panic.message);
-        }
-        for @.worries {
-            @m.push(.message);
-        }
-        @m.join("\n")
-    }
-}
 
 do {
     sub is_runtime($bt) {
         for $bt.keys {
             try {
-                return True if nqp::iseq_s($bt[$_]<sub>, 'eval')
-                    && nqp::iseq_s(
-                            nqp::join(';', $bt[$_]<sub>.get_namespace.get_name),
-                            'nqp'
-                    );
-                return False if nqp::iseq_s($bt[$_]<sub>, 'compile')
-                    && nqp::iseq_s(
-                            nqp::join(';', $bt[$_]<sub>.get_namespace.get_name),
-                            'nqp'
-                    );
+                my Mu $sub := nqp::getattr(nqp::decont($bt[$_]<sub>), ForeignCode, '$!do');
+                my Mu $codeobj := nqp::ifnull(nqp::getcodeobj($sub), Mu);
+                my $is_nqp = $codeobj && $codeobj.HOW.name($codeobj) eq 'NQPRoutine';
+                return True if nqp::iseq_s(nqp::getcodename($sub), 'eval') && $is_nqp;
+                return False if nqp::iseq_s(nqp::getcodename($sub), 'compile') && $is_nqp;
             }
         }
         return False;
@@ -195,61 +151,80 @@ do {
 
 
     sub print_exception(|) is hidden_from_backtrace {
-        my Mu $ex := nqp::atpos(pir::perl6_current_args_rpa__P(), 0);
+        my Mu $ex := nqp::atpos(nqp::p6argvmarray(), 0);
         try {
             my $e := EXCEPTION($ex);
             my Mu $err := nqp::getstderr();
 
+#?if parrot
             if $e.is-compile-time || is_runtime($ex.backtrace) {
-                $err.print: $e.gist;
-                $err.print: "\n";
+#?endif
+#?if jvm
+            if $e.is-compile-time || is_runtime(nqp::backtrace($ex)) {
+#?endif
+                nqp::printfh($err, $e.gist);
+                nqp::printfh($err, "\n");
             }
             else {
-                $err.print: "===SORRY!===\n";
-                $err.print: $ex;
-                $err.print: "\n";
+                nqp::printfh($err, "===SORRY!===\n");
+                nqp::printfh($err, $e.Str);
+                nqp::printfh($err, "\n");
             }
-            $_() for nqp::p6type(@*END_PHASERS);
+            $_() for nqp::hllize(nqp::getcurhllsym('@END_PHASERS'));
         }
         if $! {
-            pir::perl6_based_rethrow__0PP(nqp::getattr(nqp::p6decont($!), Exception, '$!ex'), $ex);
+#?if parrot
+            pir::perl6_based_rethrow__0PP(nqp::getattr(nqp::decont($!), Exception, '$!ex'), $ex);
+#?endif
+#?if !parrot
+            nqp::rethrow(nqp::getattr(nqp::decont($!), Exception, '$!ex'));
+            $ex
+#?endif
         }
     }
 
     sub print_control(|) is hidden_from_backtrace {
-        my Mu $ex := nqp::atpos(pir::perl6_current_args_rpa__P(), 0);
-        my int $type = nqp::atkey_i($ex, 'type');
-        if ($type == pir::const::CONTROL_OK) {
+        my Mu $ex := nqp::atpos(nqp::p6argvmarray(), 0);
+        my int $type = nqp::getextype($ex);
+        if ($type == nqp::const::CONTROL_WARN) {
             my Mu $err := nqp::getstderr();
-            my $msg = nqp::p6box_s(nqp::atkey_s($ex, 'message'));
-            $err.print: $msg ?? "$msg" !! "Warning";
-            $err.print: Backtrace.new($ex.backtrace, 0).nice(:oneline);
-            $err.print: "\n";
+            my $msg = nqp::p6box_s(nqp::getmessage($ex));
+            nqp::printfh($err, $msg ?? "$msg" !! "Warning");
+#?if parrot
+            nqp::printfh($err, Backtrace.new($ex.backtrace, 0).nice(:oneline));
+#?endif
+#?if jvm
+#            XXX Backtraces busted
+#            nqp::printfh($err, Backtrace.new(nqp::backtrace($ex), 0).nice(:oneline));
+#?endif
+            nqp::printfh($err, "\n");
+#?if parrot
             my $resume := nqp::atkey($ex, 'resume');
             if ($resume) {
                 $resume();
             }
+#?endif
+#?if !parrot
+            nqp::resume($ex)
+#?endif
         }
-        if ($type == pir::const::CONTROL_RETURN) {
-            die("stray return control exception");
-        }
-        if ($type == pir::const::CONTROL_LOOP_LAST) {
+        if ($type == nqp::const::CONTROL_LAST) {
             X::ControlFlow.new(illegal => 'last', enclosing => 'loop construct').throw;
         }
-        if ($type == pir::const::CONTROL_LOOP_NEXT) {
+        if ($type == nqp::const::CONTROL_NEXT) {
             X::ControlFlow.new(illegal => 'next', enclosing => 'loop construct').throw;
         }
-        if ($type == pir::const::CONTROL_LOOP_REDO) {
+        if ($type == nqp::const::CONTROL_REDO) {
             X::ControlFlow.new(illegal => 'redo', enclosing => 'loop construct').throw;
         }
-        if ($type == pir::const::CONTROL_CONTINUE) {
+        if ($type == nqp::const::CONTROL_PROCEED) {
             X::ControlFlow.new(illegal => 'proceed', enclosing => 'when clause').throw;
         }
-        if ($type == pir::const::CONTROL_BREAK) {
+        if ($type == nqp::const::CONTROL_SUCCEED) {
             # XXX: should work like leave() ?
             X::ControlFlow.new(illegal => 'succeed', enclosing => 'when clause').throw;
         }
-        if ($type == pir::const::CONTROL_TAKE) {
+        if ($type == nqp::const::CONTROL_TAKE) {
             X::ControlFlow.new(illegal => 'take', enclosing => 'gather').throw;
         }
     }
@@ -257,16 +232,26 @@ do {
     my Mu $comp := nqp::getcomp('perl6');
     $comp.HOW.add_method($comp, 'handle-exception',
         method (|) {
-            my Mu $ex := nqp::atpos(pir::perl6_current_args_rpa__P(), 1);
+            my Mu $ex := nqp::atpos(nqp::p6argvmarray(), 1);
+#?if parrot
             pir::perl6_invoke_catchhandler__vPP(&print_exception, $ex);
+#?endif
+#?if !parrot
+            print_exception($ex);
+#?endif
             nqp::exit(1);
             0;
         }
     );
     $comp.HOW.add_method($comp, 'handle-control',
         method (|) {
-            my Mu $ex := nqp::atpos(pir::perl6_current_args_rpa__P(), 1);
+            my Mu $ex := nqp::atpos(nqp::p6argvmarray(), 1);
+#?if parrot
             pir::perl6_invoke_catchhandler__vPP(&print_control, $ex);
+#?endif
+#?if !parrot
+            print_control($ex);
+#?endif
             nqp::rethrow($ex);
         }
     );
@@ -295,6 +280,22 @@ my class X::IO::Copy does X::IO is Exception {
     }
 }
 
+my class X::IO::Symlink does X::IO is Exception {
+    has $.target;
+    has $.name;
+    method message() {
+        "Failed to create symlink called '$.name' on target '$.target': $.os-error"
+    }
+}
+
+my class X::IO::Link does X::IO is Exception {
+    has $.target;
+    has $.name;
+    method message() {
+        "Failed to create link called '$.name' on target '$.target': $.os-error"
+    }
+}
+
 my class X::IO::Mkdir does X::IO is Exception {
     has $.path;
     has $.mode;
@@ -384,7 +385,7 @@ my role X::Comp is Exception {
     method sorry_heading() {
         my $color = %*ENV<RAKUDO_ERROR_COLOR> // $*OS ne 'MSWin32';
         my ($red, $clear) = $color ?? ("\e[31m", "\e[0m") !! ("", "");
-        "$red==={$clear}SORRY!$red===$clear\n"
+        "$red==={$clear}SORRY!$red===$clear Error while compiling $.filename\n"
     }
     method SET_FILE_LINE($file, $line) {
         $!filename = $file;
@@ -393,6 +394,52 @@ my role X::Comp is Exception {
     }
 }
 
+my class X::Comp::Group is Exception {
+    has $.panic;
+    has @.sorrows;
+    has @.worries;
+
+    method is-compile-time() { True }
+
+    multi method gist(::?CLASS:D:) {
+        my $r = "";
+        if $.panic || @.sorrows {
+            my $color = %*ENV<RAKUDO_ERROR_COLOR> // $*OS ne 'MSWin32';
+            my ($red, $clear) = $color ?? ("\e[31m", "\e[0m") !! ("", "");
+            $r ~= "$red==={$clear}SORRY!$red===$clear\n";
+            for @.sorrows {
+                $r ~= .gist(:!sorry, :!expect) ~ "\n";
+            }
+            if $.panic {
+                $r ~= $.panic.gist(:!sorry) ~ "\n";
+            }
+        }
+        if @.worries {
+            $r ~= $.panic || @.sorrows
+                ?? "Other potential difficulties:\n"
+                !! "Potential difficulties:\n";
+            for @.worries {
+                $r ~= .gist(:!sorry, :!expect).indent(4) ~ "\n";
+            }
+        }
+        $r
+    }
+
+    method message() {
+        my @m;
+        for @.sorrows {
+            @m.push(.message);
+        }
+        if $.panic {
+            @m.push($.panic.message);
+        }
+        for @.worries {
+            @m.push(.message);
+        }
+        @m.join("\n")
+    }
+}
+
 # XXX a hack for getting line numbers from exceptions from the metamodel
 my class X::Comp::AdHoc is X::AdHoc does X::Comp {
     method is-compile-time() { True }
@@ -407,6 +454,27 @@ my class X::NYI is Exception {
 }
 my class X::Comp::NYI is X::NYI does X::Comp { };
 
+my class X::Trait::Unknown is Exception {
+    has $.type;       # is, will, of etc.
+    has $.subtype;    # wrong subtype being tried
+    has $.declaring;  # variable, sub, parameter, etc.
+    method message () {
+        "Can't use unknown trait '$.type $.subtype' in a$.declaring declaration."
+    }
+}
+my class X::Comp::Trait::Unknown is X::Trait::Unknown does X::Comp { };
+
+my class X::Trait::NotOnNative is Exception {
+    has $.type;       # is, will, of etc.
+    has $.subtype;    # wrong subtype being tried
+    has $.native;     # type of native (optional)
+    method message () {
+        "Can't use trait '$.type $.subtype' on a native"
+          ~ ( $.native ?? " $.native." !! "." );
+    }
+}
+my class X::Comp::Trait::NotOnNative is X::Trait::NotOnNative does X::Comp { };
+
 my class X::OutOfRange is Exception {
     has $.what = 'Argument';
     has $.got = '<unknown>';
@@ -494,7 +562,7 @@ my class X::Undeclared::Symbols does X::Comp {
     }
     method message() {
         sub l(@l) {
-            my @lu = @l.uniq.sort;
+            my @lu = @l.map({ nqp::hllize($_) }).uniq.sort;
             'used at line' ~ (@lu == 1 ?? ' ' !! 's ') ~ @lu.join(', ')
         }
         sub s(@s) {
@@ -699,6 +767,15 @@ my class X::Syntax::UnlessElse does X::Syntax {
     method message() { '"unless" does not take "else", please rewrite using "if"' }
 }
 
+my class X::Syntax::KeywordAsFunction does X::Syntax {
+    has $.word;
+    has $.needparens;
+    method message {
+        "Word '$.word' interpreted as '{$.word}()' function call; please use whitespace "
+            ~ ($.needparens ?? 'around the parens' !! 'instead of parens')
+    }
+}
+
 my class X::Syntax::Malformed::Elsif does X::Syntax {
     has $.what = 'else if';
     method message() { qq{In Perl 6, please use "elsif' instead of "$.what"} }
@@ -789,8 +866,71 @@ my class X::Syntax::Missing does X::Syntax {
     method message() { "Missing $.what" }
 }
 
-my class X::Syntax::SigilWithoutName does X::Syntax {
-    method message() { 'Non-declarative sigil is missing its name' }
+my class X::Syntax::Perl5Var does X::Syntax {
+    has $.name;
+    my %m =
+      '$*'  => '^^ and $$',
+      '$"'  => '.join() method',
+      '$$'  => '$*PID',
+      '$('  => '$*GID',
+      '$)'  => '$*EGID',
+      '$<'  => '$*UID',
+      '$>'  => '$*EUID',
+      '$;'  => 'real multidimensional hashes',
+      '$&'  => '$<>',
+      '$`'  => 'explicit pattern before <(',
+      '$\'' => 'explicit pattern after )>',
+      '$,'  => '$*OUT.output_field_separator()',
+      '$.'  => "the filehandle's .line method",
+      '$\\' => "the filehandle's .ors attribute",
+      '$|'  => ':autoflush on open',
+      '$?'  => '$! for handling child errors also',
+      '$@'  => '$!',
+      '$#'  => '.fmt',
+      '$['  => 'user-defined array indices',
+      '$]'  => '$*PERL_VERSION',
+
+      '$^C' => 'COMPILING namespace',
+      '$^D' => '$*DEBUGGING',
+      '$^E' => '$!.extended_os_error',
+      '$^F' => '$*SYSTEM_FD_MAX',
+      '$^H' => '$?FOO variables',
+      '$^I' => '$*INPLACE',
+      '$^M' => 'a global form such as $*M',
+      '$^N' => '$/[*-1]',
+      '$^O' => '$?OS or $*OS',
+      '$^R' => 'an explicit result variable',
+      '$^S' => 'context function',
+      '$^T' => '$*BASETIME',
+      '$^V' => '$*PERL_VERSION',
+      '$^W' => '$*WARNING',
+      '$^X' => '$*EXECUTABLE_NAME',
+
+      '$:'  => 'Form module',
+      '$-'  => 'Form module',
+      '$+'  => 'Form module',
+      '$='  => 'Form module',
+      '$%'  => 'Form module',
+      '$^'  => 'Form module',
+      '$~'  => 'Form module',
+      '$^A' => 'Form module',
+      '$^L' => 'Form module',
+
+      '@-'  => '.from method',
+      '@+'  => '.to method',
+
+      '%-'  => '.from method',
+      '%+'  => '.to method',
+      '%^H' => '$?FOO variables',
+    ;
+    method message() {
+        my $v = $.name ~~ m/ <[ $ @ % & ]> [ \^ <[ A..Z ]> | \W ] /;
+        $v
+          ?? %m{~$v}
+            ?? "Unsupported use of $v variable; in Perl 6 please use {%m{~$v}}"
+            !! "Unsupported use of $v variable"
+          !! 'Non-declarative sigil is missing its name';
+    }
 }
 
 my class X::Syntax::Self::WithoutObject does X::Syntax {
@@ -1051,6 +1191,14 @@ my class X::Assignment::RO is Exception {
     }
 }
 
+my class X::Immutable is Exception {
+    has $.typename;
+    has $.method;
+    method message {
+        "Cannot call '$.method' on an immutable '$.typename'";
+    }
+}
+
 my class X::NoDispatcher is Exception {
     has $.redispatcher;
     method message() {
@@ -1181,6 +1329,13 @@ my class X::Numeric::Real is Exception {
     }
 }
 
+my class X::Numeric::DivideByZero is Exception {
+    has $.using;
+    method message() {
+        "Divide by zero" ~ ( $.using ?? " using $.using" !! '' );
+    }
+}
+
 my class X::PseudoPackage::InDeclaration does X::Comp {
     has $.pseudo-package;
     has $.action;
@@ -1219,6 +1374,13 @@ my class X::Multi::NoMatch is Exception {
     }
 }
 
+my class X::Caller::NotDynamic is Exception {
+    has $.symbol;
+    method message() {
+        "Cannot access '$.symbol' through CALLER, because it is not declared as dynamic";
+    }
+}
+
 {
     my %c_ex;
     %c_ex{'X::TypeCheck::Binding'} := sub ($got, $expected) is hidden_from_backtrace {
