@@ -36,18 +36,14 @@ my class Failure {
         }
     );
     method sink() { $!exception.throw unless $!handled }
-
-    # class Any has a fallback method, so we need to redefine it here
-    method postcircumfix:<{ }>(|c) { $!exception.throw }
 }
 
-
 my &fail := -> *@msg {
     my $value = @msg == 1 ?? @msg[0] !! @msg.join('');
     die $value if $*FATAL;
     try die $value;
     my $fail := Failure.new($!);
-    my Mu $return := pir::find_caller_lex__Ps('RETURN');
+    my Mu $return := nqp::getlexcaller('RETURN');
     $return($fail) unless nqp::isnull($return);
     $fail
 }
