@@ -1,4 +1,5 @@
 my class X::Bind { ... }
+my class X::Caller::NotDynamic { ... }
 
 my class PseudoStash is EnumMap {
     has Mu $!ctx;
@@ -8,7 +9,8 @@ my class PseudoStash is EnumMap {
     my int constant PICK_CHAIN_BY_NAME = 0;
     my int constant STATIC_CHAIN       = 1;
     my int constant DYNAMIC_CHAIN      = 2;
-    my int constant PRECISE_SCOPE      = 3;
+    my int constant PRECISE_SCOPE      = 4;
+    my int constant REQUIRE_DYNAMIC    = 8;
 
     method new() {
         my $obj := nqp::create(self);
@@ -27,7 +29,7 @@ my class PseudoStash is EnumMap {
                 $stash);
         },
         'CORE' => sub ($cur) {
-            my Mu $ctx := nqp::getattr(nqp::p6decont($cur), PseudoStash, '$!ctx');
+            my Mu $ctx := nqp::getattr(nqp::decont($cur), PseudoStash, '$!ctx');
             until nqp::existskey(nqp::ctxlexpad($ctx), '!CORE_MARKER') {
                 $ctx := nqp::ctxouter($ctx);
             }
@@ -41,18 +43,18 @@ my class PseudoStash is EnumMap {
         },
         'CALLER' => sub ($cur) {
             my Mu $ctx := nqp::ctxcaller(
-                nqp::getattr(nqp::p6decont($cur), PseudoStash, '$!ctx'));
+                nqp::getattr(nqp::decont($cur), PseudoStash, '$!ctx'));
             my $stash := nqp::create(PseudoStash);
             nqp::bindattr($stash, EnumMap, '$!storage', nqp::ctxlexpad($ctx));
             nqp::bindattr($stash, PseudoStash, '$!ctx', $ctx);
-            nqp::bindattr_i($stash, PseudoStash, '$!mode', PRECISE_SCOPE);
+            nqp::bindattr_i($stash, PseudoStash, '$!mode', PRECISE_SCOPE +| REQUIRE_DYNAMIC);
             nqp::setwho(
                 Metamodel::ModuleHOW.new_type(:name('CALLER')),
                 $stash);
         },
         'OUTER' => sub ($cur) {
             my Mu $ctx := nqp::ctxouter(
-                nqp::getattr(nqp::p6decont($cur), PseudoStash, '$!ctx'));
+                nqp::getattr(nqp::decont($cur), PseudoStash, '$!ctx'));
             my $stash := nqp::create(PseudoStash);
             nqp::bindattr($stash, EnumMap, '$!storage', nqp::ctxlexpad($ctx));
             nqp::bindattr($stash, PseudoStash, '$!ctx', $ctx);
@@ -69,7 +71,7 @@ my class PseudoStash is EnumMap {
                 $stash);
         },
         'UNIT' => sub ($cur) {
-            my Mu $ctx := nqp::getattr(nqp::p6decont($cur), PseudoStash, '$!ctx');
+            my Mu $ctx := nqp::getattr(nqp::decont($cur), PseudoStash, '$!ctx');
             until nqp::existskey(nqp::ctxlexpad($ctx), '!UNIT_MARKER') {
                 $ctx := nqp::ctxouter($ctx);
             }
@@ -84,7 +86,7 @@ my class PseudoStash is EnumMap {
         'SETTING' => sub ($cur) {
             # Same as UNIT, but go a little further out (two steps, for
             # internals reasons).
-            my Mu $ctx := nqp::getattr(nqp::p6decont($cur), PseudoStash, '$!ctx');
+            my Mu $ctx := nqp::getattr(nqp::decont($cur), PseudoStash, '$!ctx');
             until nqp::existskey(nqp::ctxlexpad($ctx), '!UNIT_MARKER') {
                 $ctx := nqp::ctxouter($ctx);
             }
@@ -98,73 +100,81 @@ my class PseudoStash is EnumMap {
                 $stash);
         },
         'OUR' => sub ($cur) {
-            pir::find_lex_relative__PPs(
-                nqp::getattr(nqp::p6decont($cur), PseudoStash, '$!ctx'),
+            nqp::getlexrel(
+                nqp::getattr(nqp::decont($cur), PseudoStash, '$!ctx'),
                 '$?PACKAGE')
         };
     
     method at_key($key is copy) is rw {
         $key = $key.Str;
-        if %pseudoers.exists($key) {
+        my Mu $nkey := nqp::unbox_s($key);
+        if %pseudoers.exists_key($key) {
             %pseudoers{$key}(self)
         }
-        elsif $!mode == PRECISE_SCOPE {
+        elsif nqp::bitand_i($!mode, PRECISE_SCOPE) {
             my Mu $store := nqp::getattr(self, EnumMap, '$!storage');
-            nqp::existskey($store, nqp::unbox_s($key)) ??
-                nqp::atkey($store, nqp::unbox_s($key)) !!
-                Any
+            my Mu $res := nqp::existskey($store, $nkey) ??
+                            nqp::atkey($store, $nkey) !!
+                            Any;
+            if !($res =:= Any) && nqp::bitand_i($!mode, REQUIRE_DYNAMIC) {
+                if !$res.VAR.dynamic {
+                    X::Caller::NotDynamic.new(
+                        symbol => $key,
+                    ).throw;
+                }
+            }
+            $res;
         }
-        elsif $!mode == DYNAMIC_CHAIN || $!mode == PICK_CHAIN_BY_NAME && substr($key, 1, 1) eq '*' {
-            my $found := pir::find_dynamic_lex_relative__PPs(
+        elsif nqp::bitand_i($!mode, nqp::bitor_i(DYNAMIC_CHAIN, PICK_CHAIN_BY_NAME)) && substr($key, 1, 1) eq '*' {
+            my $found := nqp::getlexreldyn(
                 nqp::getattr(self, PseudoStash, '$!ctx'),
-                nqp::unbox_s($key));
+                $nkey);
             nqp::isnull($found) ?? Any !! $found
         }
         else {
-            my $found := pir::find_lex_relative__PPs(
+            my $found := nqp::getlexrel(
                 nqp::getattr(self, PseudoStash, '$!ctx'),
-                nqp::unbox_s($key));
+                $nkey);
             nqp::isnull($found) ?? Any !! $found
         }
     }
     
     method bind_key($key is copy, \value) {
         $key = $key.Str;
-        if %pseudoers.exists($key) {
+        if %pseudoers.exists_key($key) {
             X::Bind.new(target => "pseudo-package $key").throw;
         }
-        elsif $!mode == PRECISE_SCOPE {
+        elsif nqp::bitand_i($!mode, PRECISE_SCOPE) {
             my Mu $store := nqp::getattr(self, EnumMap, '$!storage');
             nqp::bindkey($store, nqp::unbox_s($key), value)
         }
-        elsif $!mode == DYNAMIC_CHAIN || $!mode == PICK_CHAIN_BY_NAME && substr($key, 1, 1) eq '*' {
+        elsif nqp::bitand_i($!mode, nqp::bitor_i(DYNAMIC_CHAIN, PICK_CHAIN_BY_NAME)) && substr($key, 1, 1) eq '*' {
             die "Binding to dynamic variables not yet implemented";
         }
         else {
             die "This case of binding is not yet implemented";
         }
     }
-    
-    method exists($key is copy) {
+    method exists_key($key is copy) {
         $key = $key.Str;
-        if %pseudoers.exists($key) {
+        if %pseudoers.exists_key($key) {
             True
         }
-        elsif $!mode == PRECISE_SCOPE {
+        elsif nqp::bitand_i($!mode, PRECISE_SCOPE) {
             nqp::p6bool(nqp::existskey(
                 nqp::getattr(self, EnumMap, '$!storage'),
                 nqp::unbox_s($key)))
         }
-        elsif $!mode == DYNAMIC_CHAIN || $!mode == PICK_CHAIN_BY_NAME && substr($key, 1, 1) eq '*' {
+        elsif nqp::bitand_i($!mode, nqp::bitor_i(DYNAMIC_CHAIN, PICK_CHAIN_BY_NAME)) && substr($key, 1, 1) eq '*' {
             nqp::isnull(
-                pir::find_dynamic_lex_relative__PPs(
+                nqp::getlexreldyn(
                     nqp::getattr(self, PseudoStash, '$!ctx'),
                     nqp::unbox_s($key)))
                 ?? False !! True
         }
         else {
             nqp::isnull(
-                pir::find_lex_relative__PPs(
+                nqp::getlexrel(
                     nqp::getattr(self, PseudoStash, '$!ctx'),
                     nqp::unbox_s($key)))
                 ?? False !! True
