@@ -5,8 +5,10 @@ my class Rat is Cool does Rational[Int, Int] {
     multi method perl(Rat:D:) {
         my $d = $.denominator;
         return $.numerator ~ '.0' if $d == 1;
-        $d div= 5 while $d %% 5;
-        $d div= 2 while $d %% 2;
+        unless $d == 0 {
+            $d div= 5 while $d %% 5;
+            $d div= 2 while $d %% 2;
+        }
         ($d == 1) ?? self.Str !! '<' ~ $.numerator ~ '/' ~ $.denominator ~ '>';
     }
 }
@@ -33,18 +35,18 @@ sub DIVIDE_NUMBERS(Int:D \nu, Int:D \de, $t1, $t2) {
     }
     if nqp::istype($t1, FatRat) || nqp::istype($t2, FatRat) {
         my $r := nqp::create(FatRat);
-        nqp::bindattr($r, FatRat, '$!numerator',   nqp::p6decont($numerator));
-        nqp::bindattr($r, FatRat, '$!denominator', nqp::p6decont($denominator));
+        nqp::bindattr($r, FatRat, '$!numerator',   nqp::decont($numerator));
+        nqp::bindattr($r, FatRat, '$!denominator', nqp::decont($denominator));
         $r;
     } elsif $denominator < $UINT64_UPPER {
         my $r := nqp::create(Rat);
-        nqp::bindattr($r, Rat, '$!numerator',   nqp::p6decont($numerator));
-        nqp::bindattr($r, Rat, '$!denominator', nqp::p6decont($denominator));
+        nqp::bindattr($r, Rat, '$!numerator',   nqp::decont($numerator));
+        nqp::bindattr($r, Rat, '$!denominator', nqp::decont($denominator));
         $r;
     } else {
         nqp::p6box_n(nqp::div_In(
-                nqp::p6decont($numerator),
-                nqp::p6decont($denominator)
+                nqp::decont($numerator),
+                nqp::decont($denominator)
             )
         );
     }
