@@ -1,13 +1,18 @@
-my class Regex {
+my class Regex { # declared in BOOTSTRAP
+    # class Regex is Method {
+    #     has Mu $!caps;
+    #     has Mu $!nfa;
+    #     has Mu $!alt_nfas
+
     multi method ACCEPTS(Regex:D \SELF: Mu \topic) {
-        my $dollar_slash := pir::find_caller_lex_relative__PPs(
+        my $dollar_slash := nqp::getlexrelcaller(
             nqp::ctxcaller(nqp::ctxcaller(nqp::ctx())),
             '$/');
         $dollar_slash = SELF.(Cursor."!cursor_init"(topic, :c(0))).MATCH_SAVE;
     }
 
     multi method ACCEPTS(Regex:D \SELF: @a) {
-        my $dollar_slash := pir::find_caller_lex_relative__PPs(
+        my $dollar_slash := nqp::getlexrelcaller(
             nqp::ctxcaller(nqp::ctxcaller(nqp::ctx())),
             '$/');
         for @a {
@@ -17,7 +22,7 @@ my class Regex {
         Nil;
     }
     multi method ACCEPTS(Regex:D \SELF: %h) {
-        my $dollar_slash := pir::find_caller_lex_relative__PPs(
+        my $dollar_slash := nqp::getlexrelcaller(
             nqp::ctxcaller(nqp::ctxcaller(nqp::ctx())),
             '$/');
         for %h.keys {
@@ -28,10 +33,10 @@ my class Regex {
     }
 
     multi method Bool(Regex:D:) {
-        my $dollar_slash := pir::find_caller_lex_relative__PPs(
+        my $dollar_slash := nqp::getlexrelcaller(
             nqp::ctxcaller(nqp::ctxcaller(nqp::ctx())),
             '$/');
-        my $dollar_underscore := pir::find_caller_lex_relative__PPs(
+        my $dollar_underscore := nqp::getlexrelcaller(
             nqp::ctxcaller(nqp::ctxcaller(nqp::ctx())),
             '$_');
         $dollar_slash = $dollar_underscore.match(self);
