@@ -1,19 +1,28 @@
 my class Cursor {... }
 my class Range  {... }
 my class Match  {... }
-my class Buf    {... }
 my class IO::Path         { ... }
 my class X::Str::Numeric  { ... }
 my class X::Str::Match::x { ... }
 my class X::Str::Trans::IllegalKey { ... }
 my class X::Str::Trans::InvalidArg { ... }
+my class X::NYI { ... }
 
 
 my $?TABSTOP = 8;
 
-sub PARROT_ENCODING(Str:D $s) {
-    my %map = (
+sub NORMALIZE_ENCODING(Str:D $s) {
+    state %map = (
+        # fast mapping for identicals
+        'utf8'              => 'utf8',
+        'utf16'             => 'utf16',
+        'utf32'             => 'utf32',
+        'ascii'             => 'ascii',
+        'iso-8859-1'        => 'iso-8859-1',
+        # with dash
         'utf-8'             => 'utf8',
+        'utf-16'            => 'utf16',
+        'utf-32'            => 'utf32',
         # according to http://de.wikipedia.org/wiki/ISO-8859-1
         'iso_8859-1:1987'   => 'iso-8859-1',
         'iso_8859-1'        => 'iso-8859-1',
@@ -25,15 +34,17 @@ sub PARROT_ENCODING(Str:D $s) {
         'ibm819'            => 'iso-8859-1',
         'cp819'             => 'iso-8859-1',
     );
-    my Str $lc = lc($s);
-    %map{$lc} // $lc;
+    %map{$s} // %map{lc $s} // lc $s;
 }
 
-my class Str does Stringy {
+my class Str does Stringy { # declared in BOOTSTRAP
+    # class Str is Cool {
+    #     has str $!value is box_target;
+
     multi method WHICH(Str:D:) {
         nqp::box_s(
-            nqp::concat_s(
-                nqp::concat_s(nqp::unbox_s(self.^name), '|'),
+            nqp::concat(
+                nqp::concat(nqp::unbox_s(self.^name), '|'),
                 $!value
             ),
             ObjAt
@@ -45,7 +56,9 @@ my class Str does Stringy {
 
     multi method Bool(Str:D:) { self ne '' && self ne '0' }
 
-    multi method Str(Str:D:) { self }
+    multi method Str(Str:D:)     { self }
+    multi method Stringy(Str:D:) { self }
+    multi method DUMP(Str:D:) { self.perl }
 
     method Int(Str:D:) { self.Numeric.Int; }
     method Num(Str:D:) { self.Numeric.Num; }
@@ -61,11 +74,12 @@ my class Str does Stringy {
         $to_remove = 1 if $last eq "\n" || $last eq "\r";
         $to_remove = 2 if $chars > 1
             && nqp::p6box_s(nqp::substr($sself, $chars - 2)) eq "\r\n";
-        nqp::p6box_s(pir::chopn__Ssi($sself, $to_remove))
+        nqp::p6box_s(nqp::substr($sself, 0, $chars - $to_remove))
     }
 
     method chop(Str:D:) {
-        nqp::p6box_s(pir::chopn__Ssi(nqp::unbox_s(self), 1))
+        my str $sself = nqp::unbox_s(self);
+        nqp::p6box_s(nqp::substr($sself, 0, nqp::chars($sself) - 1))
     }
 
     method substr(Str:D: $start, $length? is copy) {
@@ -127,6 +141,7 @@ my class Str does Stringy {
         ~ "\x[2460,2461,2462,2463,2464,2465,2466,2467,2468,2469,246A,246B,246C,246D,246E,246F,2470,2471,2472,2473,2460]" # circled digits 1..20
         ~ "\x[2474,2475,2476,2477,2478,2479,247A,247B,247C,247D,247E,247F,2480,2481,2482,2483,2484,2485,2486,2487,2474]" # parenthesized digits 1..20
         ~ "\x[249C,249D,249E,249F,24A0,24A1,24A2,24A3,24A4,24A5,24A6,24A7,24A8,24A9,24AA,24AB,24AC,24AD,24AE,24AF,24B0,24B1,24B2,24B3,24B4,24B5,249C]" # parenthesized latin lc
+        ~ "\x[2581,2582,2583,2584,2585,2586,2587,2588]" # lower blocks
         ~ "\x[2680,2681,2682,2683,2684,2685,2680]" # die faces
         ~ "\x[2776,2777,2778,2779,277A,277B,277C,277D,277E,277F,2776]"; # dingbat negative circled 1..10
 
@@ -174,7 +189,7 @@ my class Str does Stringy {
             my int $ipos = nqp::index($RANGECHAR, $ch0);
             $ipos = $RANGECHAR.index($ch0, $ipos+1) // $ipos;
             my str $ch1 = nqp::substr($RANGECHAR, $ipos-1, 1);
-            $str = pir::replace__Ssiis($str, $r1, 1, $ch1);
+            $str = nqp::replace($str, $r1, 1, $ch1);
             # return if no carry
             return $str if $ch0 gt $ch1;
             # carry to previous position
@@ -193,12 +208,12 @@ my class Str does Stringy {
             my str $ch0  = nqp::substr($str, $r1, 1);
             my int $ipos = nqp::index($RANGECHAR, $ch0);
             my str $ch1  = nqp::substr($RANGECHAR, $ipos+1, 1);
-            $str = pir::replace__Ssiis($str, $r1, 1, $ch1);
+            $str = nqp::replace($str, $r1, 1, $ch1);
             return $str if $ch1 gt $ch0;
             # carry to previous position
             $r1 = $r1 - 1;
             # extend string if carried past first rangechar position
-            $str = pir::replace__Ssiis($str, $r0, 0,
+            $str = nqp::replace($str, $r0, 0,
                        nqp::existskey($carrydigit, $ch1) 
                            ?? nqp::atkey($carrydigit, $ch1)
                            !! $ch1)
@@ -246,7 +261,7 @@ my class Str does Stringy {
             # Handle NaN here, to make later parsing simpler
             if nqp::iseq_s(nqp::substr($str, $pos, 3), 'NaN') {
                 $pos = nqp::add_i($pos, 3);
-                return nqp::p6box_n(pir::set__Ns('NaN'));
+                return nqp::p6box_n(nqp::nan());
             }
 
             # Handle any leading +/- sign
@@ -306,9 +321,9 @@ my class Str does Stringy {
                         if nqp::iseq_i($p, -1);
                     $pos    = $p;
 
-                    my num $exp  = nqp::atpos($parse, 0);
-                    my num $coef = $frac ?? nqp::add_n($int, nqp::div_n($frac, $base)) !! $int;
-                    return nqp::p6box_n(nqp::mul_n($coef, nqp::pow_n(10, $exp)));
+                    my num $exp  = nqp::atpos($parse, 0).Num;
+                    my num $coef = $frac ?? nqp::add_n($int.Num, nqp::div_n($frac.Num, $base.Num)) !! $int.Num;
+                    return nqp::p6box_n(nqp::mul_n($coef, nqp::pow_n(10e0, $exp)));
                 }
 
                 # Multiplier with exponent, if single '*' is present
@@ -516,7 +531,7 @@ my class Str does Stringy {
             my $ch = self.substr($i, 1);
             $result ~= %esc{$ch} 
                        //  (   ((!$icu && $ch.ord >= 256)
-                               || nqp::iscclass( pir::const::CCLASS_PRINTING,
+                               || nqp::iscclass( nqp::const::CCLASS_PRINTING,
                                                   nqp::unbox_s($ch), 0))
                            ?? $ch
                            !! $ch.ord.fmt('\x[%x]')
@@ -540,11 +555,11 @@ my class Str does Stringy {
                   :continue(:$c), :pos(:$p),
                   :global(:$g), :overlap(:$ov), :exhaustive(:$ex), 
                   :st(:nd(:rd(:th(:$nth)))), :$x) {
-        my $caller_dollar_slash := pir::find_caller_lex__Ps('$/');
+        my $caller_dollar_slash := nqp::getlexcaller('$/');
         my %opts;
         if $p.defined { %opts<p> = $p }
         else { %opts<c> = $c // 0; }
-        my $patrx := $pat ~~ Code ?? $pat !! / $pat: /;
+        my $patrx := $pat ~~ Code ?? $pat !! / "$pat": /;
         my $cur := $patrx(Cursor.'!cursor_init'(self, |%opts));
 
         %opts<ov> = $ov if $ov;
@@ -618,7 +633,7 @@ my class Str does Stringy {
     multi method subst($matcher, $replacement,
                        :ii(:$samecase), :ss(:$samespace),
                        :$SET_CALLER_DOLLAR_SLASH, *%options) {
-        my $caller_dollar_slash := pir::find_caller_lex__Ps('$/');
+        my $caller_dollar_slash := nqp::getlexcaller('$/');
         my $SET_DOLLAR_SLASH     = $SET_CALLER_DOLLAR_SLASH || nqp::istype($matcher, Regex);
         my @matches              = self.match($matcher, |%options);
         try $caller_dollar_slash = $/ if $SET_DOLLAR_SLASH;
@@ -632,7 +647,9 @@ my class Str does Stringy {
             try $caller_dollar_slash = $m if $SET_DOLLAR_SLASH;
             $result ~= self.substr($prev, $m.from - $prev);
 
-            my $real_replacement = ~($replacement ~~ Callable ?? $replacement($m) !! $replacement);
+            my $real_replacement = ~($replacement ~~ Callable
+                ?? ($replacement.count == 0 ?? $replacement() !! $replacement($m))
+                !! $replacement);
             $real_replacement    = $real_replacement.samecase(~$m) if $samecase;
             $real_replacement    = $real_replacement.samespace(~$m) if $samespace;
             $result ~= $real_replacement;
@@ -662,45 +679,74 @@ my class Str does Stringy {
     }
 
     multi method split(Str:D: Regex $pat, $limit = *, :$all) {
-        return ().list if $limit ~~ Numeric && $limit <= 0;
+        return ().list
+          if $limit ~~ Numeric && $limit <= 0;
         my @matches = nqp::istype($limit, Whatever)
-                        ?? self.match($pat, :g)
-                        !! self.match($pat, :x(1..$limit-1), :g);
-        gather {
-            my $prev-pos = 0;
-            for @matches {
-                take self.substr($prev-pos, .from - $prev-pos);
-                take $_ if $all;
+          ?? self.match($pat, :g)
+          !! self.match($pat, :x(1..$limit-1), :g);
+
+        # add dummy for last
+        push @matches, Match.new( :from(self.chars) );
+        my $prev-pos = 0;
+
+        if ($all) {
+            my $elems = +@matches;
+            map {
+                my $value = self.substr($prev-pos, .from - $prev-pos);
                 $prev-pos = .to;
-            }
-            take self.substr($prev-pos);
+                # we don't want the dummy object
+                --$elems ?? ($value, $_) !! $value;
+            }, @matches;
+        }
+        else {
+            map {
+                my $value = self.substr($prev-pos, .from - $prev-pos);
+                $prev-pos = .to;
+                $value;
+            }, @matches;
         }
     }
+
     multi method split(Str:D: Cool $delimiter, $limit = *, :$all) {
         my $match-string = $delimiter.Str;
         return if self eq '' && $delimiter eq '';
+
+        my $l = $limit ~~ Whatever ?? $Inf !! $limit;
+        return ().list     if $l <= 0;
+        return (self).list if $l == 1;
+
         my $c = 0;
-        my $l = $limit ~~ Whatever ?? $Inf !! $limit - 1;
-        return ().list if $l < 0;
-        if $l >= 0 {
-            gather {
-                while $l-- > 0 {
-                    if ($match-string eq "") {
-                        last unless $c + 1 < self.chars;
-                        take self.substr($c, 1);
-                        $c++;
-                    } else {
-                        my $m = self.index($match-string, $c);
-                        last unless $m.defined;
-                        take self.substr($c, $m - $c);
-                        take $match-string if $all;
-                        $c = $m + $match-string.chars;
-                    }
+        my $done = 0;
+        if $match-string eq "" {
+            my $chars = self.chars;
+            map {
+                last if $done;
+
+                if --$chars and --$l {
+                    self.substr($c++, 1);
                 }
-                take self.substr($c);
-            }
-        } else {
-            Nil;
+                else {
+                    $done = 1;
+                    self.substr($c);
+                }
+            }, 1 .. $l;
+        }
+        else {
+            my $width = $match-string.chars;
+            map {
+                last if $done;
+
+                my $m = self.index($match-string, $c);
+                if $m.defined and --$l {
+                    my $value = self.substr($c, $m - $c);
+                    $c = $m + $width;
+                    $all ?? ($value,$match-string) !! $value;
+                }
+                else {
+                    $done = 1;
+                    self.substr($c);
+                }
+            }, 1 .. $l;
         }
     }
 
@@ -755,24 +801,18 @@ my class Str does Stringy {
     }
 
     method words(Str:D: $limit = *) {
-        self.comb( / \S+ /, $limit );
+        my @chunks := self.comb( / \S+ /, $limit );
+        +@chunks == 1 ?? @chunks[0] !! @chunks
     }
 
+    my %enc_type = utf8 => utf8, utf16 => utf16, utf32 => utf32;
     method encode(Str:D $encoding = 'utf8') {
-        my $buf := Buf.new;
-        my $bb := pir::new__Ps('ByteBuffer');
-        pir::set__vPS($bb, pir::trans_encoding__SSI(
-            nqp::unbox_s(self),
-            pir::find_encoding__IS(nqp::unbox_s(PARROT_ENCODING($encoding)))
-        ));
-        nqp::bindattr_s($buf, Buf, '$!buffer', $bb.get_string('binary'));
-        $buf;
+        my $enc      := NORMALIZE_ENCODING($encoding);
+        my $enc_type := %enc_type.exists_key($enc) ?? %enc_type{$enc} !! blob8;
+        nqp::encode(nqp::unbox_s(self), nqp::unbox_s($enc), nqp::decont($enc_type.new))
     }
 
-    method capitalize(Str:D:) is DEPRECATED {
-        self.subst(:g, rx/\w+/, -> $_ { .Str.lc.ucfirst });
-    }
-    method wordcase(Str:D: :&filter = &tclc, :$where = True) {
+    method wordcase(Str:D: :&filter = &tclc, Mu :$where = True) {
         self.subst(:g, / [<:L> \w* ] +% <['\-]> /, -> $m {
             my Str $s = $m.Str;
             $s ~~ $where ?? filter($s) !! $s;
@@ -960,9 +1000,8 @@ my class Str does Stringy {
         my Int $outdent = $steps ~~ Whatever ?? $common-prefix
                                              !! -$steps;
 
-        warn sprintf('Asked to remove %d spaces, ' ~
-                     'but the shortest indent is %d spaces',
-                     $outdent, $common-prefix) if $outdent > $common-prefix;
+        warn "Asked to remove $outdent spaces, but the shortest indent is $common-prefix spaces"
+            if $outdent > $common-prefix;
 
         # Work backwards from the right end of the indent whitespace, removing
         # array elements up to # (or over, in the case of tab-explosion)
@@ -980,10 +1019,6 @@ my class Str does Stringy {
         nqp::p6box_i(nqp::chars(nqp::unbox_s(self)))
     }
 
-    method tclc(Str:D:) returns Str:D {
-        nqp::p6box_s(nqp::tclc(nqp::unbox_s(self)))
-    }
-
     method path(Str:D:) returns IO::Path:D {
         IO::Path.new(self)
     }
@@ -994,9 +1029,9 @@ multi prefix:<~>(Str:D \a)  returns Str:D { a }
 multi prefix:<~>(str $a)    returns str   { $a }
 
 multi infix:<~>(Str:D \a, Str:D \b) returns Str:D {
-    nqp::p6box_s(nqp::concat_s(nqp::unbox_s(a), nqp::unbox_s(b)))
+    nqp::p6box_s(nqp::concat(nqp::unbox_s(a), nqp::unbox_s(b)))
 }
-multi infix:<~>(str $a, str $b) returns str { nqp::concat_s($a, $b) }
+multi infix:<~>(str $a, str $b) returns str { nqp::concat($a, $b) }
 
 multi infix:<x>(Str:D $s, Int:D $repetition) returns Str:D {
     $repetition < 0
@@ -1082,6 +1117,20 @@ multi prefix:<~^>(Str \a) {
     fail "prefix:<~^> NYI";   # XXX
 }
 
+# XXX: String-wise shifts NYI
+multi infix:«~>»(Str:D \a, Int:D \b) returns Str:D {
+    X::NYI.new(feature => "infix:«~>»").throw;
+}
+multi infix:«~>»(str $a, int $b) {
+    X::NYI.new(feature => "infix:«~>»").throw;
+}
+multi infix:«~<»(Str:D \a, Int:D \b) returns Str:D {
+    X::NYI.new(feature => "infix:«~<»").throw;
+}
+multi infix:«~<»(str $a, int $b) {
+    X::NYI.new(feature => "infix:«~<»").throw;
+}
+
 multi sub ords(Str $s) returns List:D {
     my Int $c  = $s.chars;
     my str $ns = nqp::unbox_s($s);
@@ -1141,7 +1190,3 @@ sub substr-rw($s is rw, $from = 0, $chars = $s.chars - $from) {
         }
     );
 }
-
-multi sub tclc(Str:D $s) returns Str:D {
-    nqp::p6box_s(nqp::tclc(nqp::unbox_s($s)));
-}
