@@ -149,42 +149,18 @@ sub default-formatter(DateTime $dt, Bool :$subseconds) {
          !! 'Z';
 }
 
-my class DateTime-local-timezone does Callable {
-    multi method Str(DateTime-local-timezone:D:) { '<local time zone>' }
-    multi method perl(DateTime-local-timezone:D:) { '$*TZ' }
-
-    method postcircumfix:<( )>($args) { self.offset(|$args) }
-
-    method offset(DateTime:D $dt, $to-utc) {
-        # We construct local and UTC DateTimes, calculate POSIX times
-        # (pretending the local DateTime is actually in UTC), and
-        # return the difference. Surprisingly, this actually works!
-        if $to-utc {
-            my Mu $fia := pir::new__PS('FixedIntegerArray');
-            pir::set__vPI($fia, 9);
-            nqp::bindpos($fia, 0, nqp::unbox_i($dt.whole-second));
-            nqp::bindpos($fia, 1, nqp::unbox_i($dt.minute));
-            nqp::bindpos($fia, 2, nqp::unbox_i($dt.hour));
-            nqp::bindpos($fia, 3, nqp::unbox_i($dt.day));
-            nqp::bindpos($fia, 4, nqp::unbox_i($dt.month));
-            nqp::bindpos($fia, 5, nqp::unbox_i($dt.year));
-            nqp::bindpos($fia, 8, -1);
-            nqp::p6box_i(pir::encodelocaltime__IP($fia)) - $dt.posix(True);
-        } else {
-            my $p = $dt.posix;
-            my ($year, $month, $day, $hour, $minute, $second);
-            my Mu $fia := pir::decodelocaltime__PI(nqp::unbox_i($p.Int));
-            $second = nqp::p6box_i(nqp::atpos_i($fia, 0));
-            $minute = nqp::p6box_i(nqp::atpos_i($fia, 1));
-            $hour   = nqp::p6box_i(nqp::atpos_i($fia, 2));
-            $day    = nqp::p6box_i(nqp::atpos_i($fia, 3));
-            $month  = nqp::p6box_i(nqp::atpos_i($fia, 4));
-            $year   = nqp::p6box_i(nqp::atpos_i($fia, 5));
-            DateTime\
-                .new(:$year, :$month, :$day, :$hour, :$minute, :$second)\
-                .posix - $p;
-        }
-    }
+sub get-local-timezone-offset {
+  my $utc = DateTime.new(now).posix.Int;
+  my Mu $fia := nqp::p6decodelocaltime(nqp::unbox_i($utc));
+  my $second = nqp::p6box_i(nqp::atpos_i($fia, 0));
+  my $minute = nqp::p6box_i(nqp::atpos_i($fia, 1));
+  my $hour   = nqp::p6box_i(nqp::atpos_i($fia, 2));
+  my $day    = nqp::p6box_i(nqp::atpos_i($fia, 3));
+  my $month  = nqp::p6box_i(nqp::atpos_i($fia, 4));
+  my $year   = nqp::p6box_i(nqp::atpos_i($fia, 5));
+  my $local  = DateTime.new(:$year, :$month, :$day, :$hour, :$minute, :$second);
+  my $ltime  = $local.posix(True).Int;
+  $ltime - $utc;
 }
 
 my class DateTime does Dateish {
@@ -197,13 +173,16 @@ my class DateTime does Dateish {
      has     $.second    = 0.0;
      has     $.timezone  = 0; # UTC
      has     &.formatter = &default-formatter;
-     has Int $!saved-offset;
        # Not an optimization but a necessity to ensure that
        # $dt.utc.local.utc is equivalent to $dt.utc. Otherwise,
        # DST-induced ambiguity could ruin our day.
+    
+     multi method new() {
+        fail "Must provide arguments to DateTime.new()";
+    }
  
     multi method new(Int :$year!, :&formatter=&default-formatter, *%_) {
-        my $dt = self.bless(*, :$year, :&formatter, |%_);
+        my $dt = self.bless(:$year, :&formatter, |%_);
         $dt.check-date;
         $dt.check-time;
         $dt;
@@ -272,7 +251,7 @@ my class DateTime does Dateish {
         my $day   = $e - (153 * $m + 2) div 5 + 1;
         my $month = $m + 3 - 12 * ($m div 10);
         my $year  = $b * 100 + $d - 4800 + $m div 10;
-        self.bless(*, :$year, :$month, :$day,
+        self.bless(:$year, :$month, :$day,
             :$hour, :$minute, :$second,
             :&formatter).in-timezone($timezone);
     }
@@ -296,9 +275,9 @@ my class DateTime does Dateish {
             if $6 eq 'Z' {
                 $timezone = 0;                
             } else {
-                $timezone = (($6[0][1]*60 + $6[0][2]) * 60).Int;
+                $timezone = (($6[1]*60 + $6[2]) * 60).Int;
                   # RAKUDO: .Int is needed to avoid to avoid the nasty '-0'.
-                $6[0][0] eq '-' and $timezone = -$timezone;
+                $6[0] eq '-' and $timezone = -$timezone;
             }
         }
         self.new(:$year, :$month, :$day, :$hour, :$minute,
@@ -318,7 +297,7 @@ my class DateTime does Dateish {
     method clone-without-validating(*%_) { # A premature optimization.
         my %args = { :$!year, :$!month, :$!day, :$!hour, :$!minute,
                      :$!second, :$!timezone, :&!formatter, %_ };
-        self.bless(*, |%args);
+        self.bless(|%args);
     }
 
     method Instant() {
@@ -339,10 +318,15 @@ my class DateTime does Dateish {
     }
 
     method offset {
-        $!saved-offset or
-            $!timezone ~~ Callable
-         ?? $!timezone(self, True)
-         !! $!timezone
+        $!timezone.Int;
+    }
+
+    method offset-in-minutes {
+        $!timezone.Int / 60;
+    }
+
+    method offset-in-hours {
+        $!timezone.Int / 60 / 60;
     }
 
     method delta($amount, TimeUnit $unit) {
@@ -427,9 +411,7 @@ my class DateTime does Dateish {
     method in-timezone($timezone) {
         $timezone eqv $!timezone and return self;
         my $old-offset = self.offset;
-        my $new-offset = $timezone ~~ Callable
-          ?? $timezone(self.utc, False)
-          !! $timezone;
+        my $new-offset = $timezone.Int;
         my %parts;
         # Is the logic for handling leap seconds right?
         # I don't know, but it passes the tests!
@@ -445,7 +427,7 @@ my class DateTime does Dateish {
            self.ymd-from-daycount\
                (self.get-daycount + floor $c / 24);
         self.clone-without-validating:
-            :$timezone, saved-offset => $new-offset, |%parts;
+            :$timezone, |%parts;
     }
 
     method utc() {
@@ -469,11 +451,13 @@ my class DateTime does Dateish {
             second => $.second.perl,
             (timezone => $.timezone.perl
                 unless $.timezone === 0),
-            (:$!saved-offset
-                if $!saved-offset and $.timezone ~~ Callable),
             (formatter => $.formatter.perl
                 unless &.formatter eqv &default-formatter)
     }
+
+    multi method gist(DateTime:D:) {
+            self.Str;
+    }
 }
 
 my class Date does Dateish {
@@ -488,7 +472,7 @@ my class Date does Dateish {
     method get-daycount { $!daycount }
 
     multi method new(:$year!, :$month = 1, :$day = 1) {
-        my $d = self.bless(*, :$year, :$month, :$day);
+        my $d = self.bless(:$year, :$month, :$day);
         $d.check-date;
         $d!set-daycount(self.daycount-from-ymd($year,$month,$day));
         $d;
@@ -517,7 +501,7 @@ my class Date does Dateish {
     }
 
     multi method new(DateTime $dt) {
-        self.bless(*, 
+        self.bless(
             :year($dt.year), :month($dt.month), :day($dt.day),
             :daycount(self.daycount-from-ymd($dt.year,$dt.month,$dt.day))
         );
@@ -525,8 +509,8 @@ my class Date does Dateish {
 
     multi method WHICH(Date:D:) {
         nqp::box_s(
-            nqp::concat_s(
-                nqp::concat_s(nqp::unbox_s(self.^name), '|'),
+            nqp::concat(
+                nqp::concat(nqp::unbox_s(self.^name), '|'),
                 nqp::unbox_i($!daycount)
             ),
             ObjAt
@@ -535,7 +519,7 @@ my class Date does Dateish {
 
     method new-from-daycount($daycount) {
         my ($year, $month, $day) = self.ymd-from-daycount($daycount);
-        self.bless(*, :$daycount, :$year, :$month, :$day);
+        self.bless(:$daycount, :$year, :$month, :$day);
     }
 
     method today() {
@@ -637,7 +621,36 @@ multi infix:«>»(Date:D $a, Date:D $b) {
     $a.daycount > $b.daycount
 }
 
-$PROCESS::TZ = DateTime-local-timezone.new;
+$PROCESS::TZ = get-local-timezone-offset();
+
+sub sleep($seconds = Inf --> Nil) {
+    if $seconds ~~ (Inf|Whatever) {
+        nqp::sleep(1e16) while True;
+    }
+    elsif $seconds > 0 {
+        nqp::sleep($seconds.Num);
+    }
+    Nil;
+}
+
+sub sleep-timer (Real $seconds = Inf --> Duration) {
+    if $seconds <= 0 {
+        Duration.new(0);
+    }
+    else {
+        my $time1 = now;
+        nqp::sleep($seconds.Num);
+        Duration.new( ( $seconds - now - $time1 ) max 0 );
+    }
+}
+
+sub sleep-till (Instant $till --> Bool) {
+    my $seconds = $till - now;
+    return False if $seconds < 0;
+
+    1 while $seconds = sleep-timer($seconds);
+    True;
+}
 
 # =begin pod
 # 
